from pydantic import BaseModel
from textual.screen import Screen
from textual.binding import Binding
# > Typing
from typing import Optional, Generator, Type, Any
# > Local Import's
from ..seaplayer import SeaPlayer
from ..codeсbase import CodecBase
from .pluginloader import PluginLoader

# ! Plugin Info Class
class PluginInfo(BaseModel):
    name: str
    name_id: str
    version: str
    author: str
    description: Optional[str]=None
    url: Optional[str]=None

# ! Plugin Base Class
class PluginBase:
    app: SeaPlayer
    pl: PluginLoader
    info: PluginInfo
    
    def __init_repr__(self) -> str: ...
    def __init__(self, app: SeaPlayer, pl: PluginLoader, info: PluginInfo) -> None: ...
    # ! App Specific Methods
    def install_screen(self, name: str, screen: Screen) -> None: ...
    def add_codecs(self, *codecs: Type[CodecBase]) -> None: ...
    def on_bindings(self) -> Generator[Binding, Any, None]: ...
    # ! App On Methods
    def on_init(self) -> None: ...
    def on_run(self) -> None: ...
    async def on_compose(self) -> None: ...
    async def on_quit(self) -> None: ...



import os
import pickle
# > Typing
from typing import TypeVar, Any

# ! T
D = TypeVar("D")

# ! Main Class
class Cacher:
    def __init__(self, cache_dirpath: str) -> None:
        self.main_dirpath = os.path.abspath(cache_dirpath)
        self.vars_dirpath = os.path.join(self.main_dirpath, "vars")
        # * Create Directory
        os.makedirs(self.main_dirpath, 0o755, True)
        os.makedirs(self.vars_dirpath, 0o755, True)
        # * Check Directory
        assert os.path.isdir(self.main_dirpath)
        assert os.path.isdir(self.vars_dirpath)
    
    # ! I/O Methods
    def write(self, data: Any, filepath: str) -> None:
        with open(filepath, "wb") as file:
            pickle.dump(data, file)
    
    def read(self, filepath: str, default: D) -> D:
        try:
            with open(filepath, "rb") as file:
                return pickle.load(file)
        except:
            self.write(default, filepath)
            return default
    
    # ! Var Methods
    def write_var(self, value: Any, name: str, *, group: str="main") -> None:
        self.write(value, os.path.join(self.vars_dirpath, f"{name}-{group}.pycache"))
    
    def read_var(self, name: str, default: D, *, group: str="main") -> D:
        return self.read(os.path.join(self.vars_dirpath, f"{name}-{group}.pycache"), default)
    
    # ! Main Methods
    def var(self, name: str, default: D, *, group: str="main") -> D:
        def setter(s, value: D) -> None:
            self.write_var(value, name, group=group)
        def getter(s) -> D:
            return self.read_var(name, default, group=group)
        return property(getter, setter)



import sys
import subprocess

# ! Main Class
class PIPManager:
    def __init__(self) -> None:
        self.python_path = sys.executable
    
    def __call__(self, *args: str) -> str:
        return subprocess.check_output([self.python_path, "-m", "pip", *args]).decode(errors="ignore")
    
    def install(self, *args: str) -> str:
        return self.__call__("install", *args)
    
    def install_requirements(self, filepath: str, upgrade: bool=False) -> str:
        cmd = ["-U", "-r", filepath] if upgrade else ["-r", filepath]
        return self.install(*cmd)

# ! Initial
pip = PIPManager()


from rich.console import Console
from textual.widgets import RichLog
# > Typing
from typing import TypeVar
# > Local Import's
from ..functions import rich_exception

# ! Vars
console = Console()

# ! Types
RETURN = TypeVar('RETURN')

# ! Main Class
class LogMenu(RichLog):
    def __init__(self, chap_max_width: int=8, enable_logging: bool=True, **kwargs):
        self.enable_logging = enable_logging
        self.chap_max_width = chap_max_width
        
        if kwargs.get("classes", None) is not None:
            kwargs["classes"] = kwargs["classes"] + " log-menu -hidden"
        else:
            kwargs["classes"] = "log-menu -hidden"
        
        super().__init__(**kwargs)
    
    def write_log(self, chap: str, msg: str, *, chap_color: str="green", in_console: bool=False) -> None:
        if self.enable_logging:
            text = f"[[{chap_color}]{chap.center(self.chap_max_width)}[/]]: {msg}"
            self.write(text, shrink=False)
            if in_console:
                console.print(text)
    
    def info(self, msg: str, *, in_console: bool=False) -> None:
        self.write_log("INFO", msg, chap_color="green", in_console=in_console)
    
    def error(self, msg: str, *, in_console: bool=False) -> None:
        self.write_log("ERROR", msg, chap_color="red", in_console=in_console)
    
    def warn(self, msg: str, *, in_console: bool=False) -> None:
        self.write_log("WARN", msg, chap_color="orange", in_console=in_console)
    
    def exception(self, e: Exception, *, in_console: bool=False) -> None:
        self.write_log("ERROR", rich_exception(e), chap_color="red", in_console=in_console)



from .MP3 import MP3Codec
from .OGG import OGGCodec
from .WAV import WAVECodec
from .MIDI import MIDICodec
from .FLAC import FLACCodec
from .URLS import URLSoundCodec
from .Any import AnyCodec
from ..codeсbase import CodecBase
from typing import List

codecs: List[CodecBase] = [ MP3Codec, OGGCodec, WAVECodec, MIDICodec, FLACCodec, URLSoundCodec, AnyCodec ]
